from pm4py.algo.conformance.alignments.versions.state_equation_a_star import apply
from pm4py.algo.conformance.alignments.versions.state_equation_a_star import apply_trace_net

from proved.artifacts.behavior_graph import behavior_graph
from proved.artifacts.behavior_net import behavior_net as behavior_net_builder
from proved.artifacts.behavior_net.utils import acyclic_net_variants


def alignment_bounds_su_log(log, petri_net, initial_marking, final_marking, parameters=None):
    """
    Returns the lower and upper bounds for conformance of a strongly uncertain log against a reference Petri net.

    :param log: the strongly uncertain event log
    :param petri_net: the reference Petri net
    :param initial_marking: the initial marking of the reference Petri net
    :param final_marking: the final marking of the reference Petri net
    :param parameters: the optional parameters for alignments
    :return: a list of 2-tuples containing the alignment results for the upper and lower bounds for conformance of the traces in the log
    """

    return [alignment_bounds_su_trace(trace, petri_net, initial_marking, final_marking, parameters) for trace in log]


def alignment_bounds_su_trace(trace, petri_net, initial_marking, final_marking, parameters=None):
    """
    Returns the lower and upper bounds for conformance of a strongly uncertain trace against a reference Petri net by aligning all possible realizations.

    :param trace: the strongly uncertain trace
    :param petri_net: the reference Petri net
    :param initial_marking: the initial marking of the reference Petri net
    :param final_marking: the final marking of the reference Petri net
    :param parameters: the optional parameters for alignments
    :return: a 2-tuple containing the alignment results for the upper and lower bounds for conformance of the trace
    """

    # Obtains the behavior net of the trace
    behavior_net = behavior_net_builder.BehaviorNet(behavior_graph.BehaviorGraph(trace))
    align_lower_bound = alignment_lower_bound_su_trace(behavior_net, behavior_net.initial_marking, behavior_net.final_marking, petri_net, initial_marking, final_marking, parameters)
    align_upper_bound_real_size = alignment_upper_bound_su_trace_bruteforce(behavior_net, behavior_net.initial_marking, behavior_net.final_marking, petri_net, initial_marking, final_marking, parameters)

    return align_lower_bound, align_upper_bound_real_size[0], align_upper_bound_real_size[1]


def alignment_upper_bound_su_trace_bruteforce(behavior_net, bn_i, bn_f, petri_net, initial_marking, final_marking, parameters=None):
    """
    Returns the upper bound for conformance of a strongly uncertain trace against a reference Petri net by aligning all possible realizations.

    :param behavior_net: the behavior net of a strongly uncertain trace
    :param bn_i: the initial marking of the behavior net
    :param bn_f: the final marking of the behavior net
    :param petri_net: the reference Petri net
    :param initial_marking: the initial marking of the reference Petri net
    :param final_marking: the final marking of the reference Petri net
    :param parameters: the optional parameters for alignments
    :return: the alignment results for the upper bound for conformance of the trace
    """

    # Obtains all the realizations of the trace by executing all possible variants from the behavior net
    r_s = acyclic_net_variants(behavior_net, bn_i, bn_f)

    # Computes the upper bound for conformance via bruteforce on the realization set
    alignments = [apply(trace, petri_net, initial_marking, final_marking, parameters) for trace in r_s]

    return max(alignments, key=lambda x: x['cost']), len(r_s)


def alignment_lower_bound_su_trace(behavior_net, bn_i, bn_f, petri_net, initial_marking, final_marking, parameters=None):
    """
    Returns the lower bound for conformance of a strongly uncertain trace against a reference Petri net by aligning using the product between the reference Petri net and the behavior net of the trace.

    :param behavior_net: the behavior net of a strongly uncertain trace
    :param bn_i: the initial marking of the behavior net
    :param bn_f: the final marking of the behavior net
    :param petri_net: the reference Petri net
    :param initial_marking: the initial marking of the reference Petri net
    :param final_marking: the final marking of the reference Petri net
    :param parameters: the optional parameters for alignments
    :return: the alignment results for the lower bound for conformance of the trace
    """

    return apply_trace_net(petri_net, initial_marking, final_marking, behavior_net, bn_i, bn_f, parameters)


def alignment_lower_bound_su_trace_bruteforce(behavior_net, bn_i, bn_f, petri_net, initial_marking, final_marking, parameters=None):
    """
    Returns the lower bound for conformance of a strongly uncertain trace against a reference Petri net by aligning all possible realizations.

    :param behavior_net: the behavior net of a strongly uncertain trace
    :param bn_i: the initial marking of the behavior net
    :param bn_f: the final marking of the behavior net
    :param petri_net: the reference Petri net
    :param initial_marking: the initial marking of the reference Petri net
    :param final_marking: the final marking of the reference Petri net
    :param parameters: the optional parameters for alignments
    :return: the alignment results for the upper bound for conformance of the trace
    """

    # Obtains all the realizations of the trace by executing all possible variants from the behavior net
    r_s = acyclic_net_variants(behavior_net, bn_i, bn_f)

    # Computes the lower bound for conformance via bruteforce on the realization set
    alignments = [apply(trace, petri_net, initial_marking, final_marking, parameters) for trace in r_s]

    return min(alignments, key=lambda x: x['cost'])



import string

from graphviz import Digraph


def behavior_graph_graphviz(behavior_graph, name='bg', format='png', filename='bg', rankdir='LR'):
    """
    Returns a Graphviz visualization from a behavior graph

    :param behavior_graph: A behavior graph
    :type behavior_graph:
    :param name: The name of the Graphviz object
    :type name:
    :param format: The format of the graphical object
    :type format:
    :param filename: The name of the filename for export from Graphviz
    :type filename:
    :param rankdir: The direction in which to draw the graph
    :type rankdir:
    :return: The Graphviz object of the behavior graph in input
    :rtype:
    """

    g = Digraph(name, format=format, filename=filename + '.' + format)
    g.attr(rankdir=rankdir)
    for bg_node in behavior_graph.nodes:
        if None in bg_node[1]:
            g.attr('node', style='dashed')
        else:
            g.attr('node', style='solid')
        g.node(''.join([act.translate(str.maketrans('', '', string.punctuation + ' ')) for act in bg_node[1] if act is not None]) + str(bg_node[0]), label=', '.join([act for act in bg_node[1] if act is not None]))
    for bg_node1, bg_node2 in behavior_graph.edges:
        g.edge(''.join([act.translate(str.maketrans('', '', string.punctuation + ' ')) for act in bg_node1[1] if act is not None]) + str(bg_node1[0]), ''.join([act.translate(str.maketrans('', '', string.punctuation + ' ')) for act in bg_node2[1] if act is not None]) + str(bg_node2[0]))
    return g



from pm4py.objects import petri


class BehaviorNet(petri.petrinet.PetriNet):
    """
    Class that represents a behavior net, a sound workflow Petri net that can replay all realizations of an uncertain trace.
    For more information refer to:
        Pegoraro, Marco, and Wil MP van der Aalst. "Mining uncertain event data in process mining." 2019 International Conference on Process Mining (ICPM). IEEE, 2019.
    """

    def __init__(self, behavior_graph):
        petri.petrinet.PetriNet.__init__(self)

        # Creating sink and source place, and invisible transitions connecting them to the rest of the net
        source_place = petri.petrinet.PetriNet.Place('source')
        self.places.add(source_place)
        source_trans = petri.petrinet.PetriNet.Transition('t_source', None)
        self.transitions.add(source_trans)
        petri.utils.add_arc_from_to(source_place, source_trans, self)

        sink_place = petri.petrinet.PetriNet.Place('sink')
        self.places.add(sink_place)
        sink_trans = petri.petrinet.PetriNet.Transition('t_sink', None)
        self.transitions.add(sink_trans)
        petri.utils.add_arc_from_to(sink_trans, sink_place, self)

        # Creating transitions for each node in the graph
        node_trans = {}
        for i, node in enumerate(behavior_graph.nodes):
            transition_set = {petri.petrinet.PetriNet.Transition('t' + str(i) + '_' + str(activity_label), activity_label) for activity_label in node[1]}
            node_trans[id(node)] = transition_set
            for transition in transition_set:
                self.transitions.add(transition)

        for i, node_from in enumerate(behavior_graph.nodes):
            # Each activity that can start the trace have to be connected through an AND-split to the starting invisible transition
            if not next(behavior_graph.predecessors(node_from), None):
                place_from_source = petri.petrinet.PetriNet.Place('source_to_' + str(id(node_from)))
                for transition in node_trans[id(node_from)]:
                    self.places.add(place_from_source)
                    petri.utils.add_arc_from_to(place_from_source, transition, self)
                petri.utils.add_arc_from_to(source_trans, place_from_source, self)

            # Every arc in the behavior graph is translated to a place in the behavior net, describing the precedence relationship between nodes
            # For each successor of the current node, all the transitions of the current node are connected to all the transitions in the successor through a place
            for node_to in behavior_graph.successors(node_from):
                place = petri.petrinet.PetriNet.Place(str(id(node_from)) + str(id(node_to)))
                self.places.add(place)
                for transition in node_trans[id(node_from)]:
                    petri.utils.add_arc_from_to(transition, place, self)
                for transition in node_trans[id(node_to)]:
                    petri.utils.add_arc_from_to(place, transition, self)

            # Each activity that can end the trace have to be connected through an AND-join to the ending invisible transition
            if not next(behavior_graph.successors(node_from), None):
                place_to_sink = petri.petrinet.PetriNet.Place(str(id(node_from)) + '_to_sink')
                for transition in node_trans[id(node_from)]:
                    self.places.add(place_to_sink)
                    petri.utils.add_arc_from_to(transition, place_to_sink, self)
                petri.utils.add_arc_from_to(place_to_sink, sink_trans, self)

        # Initial and final markings are just one token in the source place and one token in the sink place, respectively
        self.initial_marking = petri.petrinet.Marking({source_place: 1})
        self.final_marking = petri.petrinet.Marking({sink_place: 1})

    def __set_initial_marking(self, initial_marking):
        self.__initial_marking = initial_marking

    def __get_initial_marking(self):
        return self.__initial_marking

    def __set_final_marking(self, final_marking):
        self.__final_marking = final_marking

    def __get_final_marking(self):
        return self.__final_marking

    initial_marking = property(__get_initial_marking, __set_initial_marking)
    final_marking = property(__get_final_marking, __set_final_marking)



from pm4py.objects import petri
from pm4py.objects.log.log import Trace, Event
from pm4py.util.xes_constants import DEFAULT_NAME_KEY


def acyclic_net_variants(net, initial_marking, final_marking, activity_key=DEFAULT_NAME_KEY):
    """
    Given an acyclic accepting Petri net, initial and final marking extracts a set of variants (in form of traces)
    replayable on the net.
    Warning: this function is based on a marking exploration. If the accepting Petri net contains loops, the method
    will not work properly as it stops the search if a specific marking has already been encountered.

    Parameters
    ----------
    :param net: An acyclic workflow net
    :param initial_marking: The initial marking of the net.
    :param final_marking: The final marking of the net.

    Returns
    -------
    :return: variants: :class:`list` Set of variants - in the form of Trace objects - obtainable executing the net

    """
    active = {(initial_marking, ())}
    visited = set()
    variants = set()
    hash_final_marking = hash(final_marking)
    while active:
        curr_marking, curr_partial_trace = active.pop()
        hash_curr_pair = hash((curr_marking, curr_partial_trace))
        enabled_transitions = petri.semantics.enabled_transitions(net, curr_marking)
        for transition in enabled_transitions:
            if transition.label is not None:
                next_partial_trace = curr_partial_trace + (repr(transition),)
            else:
                next_partial_trace = curr_partial_trace
            next_marking = petri.semantics.execute(transition, net, curr_marking)
            hash_next_pair = hash((next_marking, next_partial_trace))

            if hash(next_marking) == hash_final_marking:
                variants.add(next_partial_trace)
            else:
                # If the next marking hash is not in visited, if the next marking+partial trace is different from the current one+partial trace
                if hash_next_pair not in visited and hash_curr_pair != hash_next_pair:
                    active.add((next_marking, next_partial_trace))
        visited.add(hash_curr_pair)
    trace_variants = []
    for variant in variants:
        trace = Trace()
        for activity_label in variant:
            trace.append(Event({activity_key: activity_label}))
        trace_variants.append(trace)
    return trace_variants



from random import random, sample
from copy import copy
from datetime import timedelta

from pm4py.objects.log.util.xes import DEFAULT_TIMESTAMP_KEY

from proved.xes_keys import DEFAULT_U_TIMESTAMP_MIN_KEY, DEFAULT_U_TIMESTAMP_MAX_KEY


def add_uncertain_timestamp_to_log(p, log=None, log_map=None, timestamp_key=DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=DEFAULT_U_TIMESTAMP_MAX_KEY):
    """
    Adds possible activity labels to events in a trace with a certain probability, up to a maximum.

    :param trace: the trace
    :param p: the probability of overlapping timestamps with previous events
    :param p_right: the probability of overlapping timestamps with successive events
    :param max_overlap_left: the maximum number of events that a timestamp can overlap
    :param max_overlap_right: the maximum number of events that a timestamp can overlap
    :param timestamp_key: the xes key for the timestamp
    :param u_timestamp_min_key: the xes key for the minimum value of an uncertain timestamp
    :param u_timestamp_max_key: the xes key for the maximum value of an uncertain timestamp
    :return:
    """

    if p > 0.0:
        if log_map is None:
            if log is None:
                raise ValueError('Parameters log and log_map cannot both be None.')
            else:
                log_map = {}
                i = 0
                for trace in log:
                    for j in range(len(trace)):
                        log_map[i] = (trace, j)
                        i += 1

        to_alter = max(0, round(len(log_map) * p))
        indices_to_alter = sample(frozenset(log_map), to_alter)
        for i in indices_to_alter:
            trace, j = log_map[i]
            # trace[j][u_timestamp_min_key] = copy(min(trace[j][timestamp_key], trace[max(j - 1, 0)][timestamp_key])) - timedelta(milliseconds=100)
            # trace[j][u_timestamp_max_key] = copy(max(trace[j][timestamp_key], trace[min(j + 1, len(trace) - 1)][timestamp_key])) + timedelta(milliseconds=100)
            # If the event is the first in the trace, we alter the timestamp to overlap the following event. If the event is the last in the trace, we alter the
            # timestamp to overlap the previous. In any other case, we pick either one at random.
            if j == 0 or (j != len(trace) - 1 and random() < .5):
                trace[j][u_timestamp_min_key] = copy(trace[j][timestamp_key]) - timedelta(milliseconds=100)
                trace[j][u_timestamp_max_key] = copy(max(trace[j][timestamp_key], trace[min(j + 1, len(trace) - 1)][timestamp_key])) + timedelta(milliseconds=100)
            else:
                trace[j][u_timestamp_min_key] = copy(min(trace[j][timestamp_key], trace[max(j - 1, 0)][timestamp_key])) - timedelta(milliseconds=100)
                trace[j][u_timestamp_max_key] = copy(trace[j][timestamp_key]) + timedelta(milliseconds=100)


def add_uncertain_timestamp_to_log_montecarlo(log, p_left, p_right, max_overlap_left=0, max_overlap_right=0, timestamp_key=DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=DEFAULT_U_TIMESTAMP_MAX_KEY):
    """
    Adds possible activity labels to events in an event log with a certain probability, up to a maximum.

    :param log: the event log
    :param p_left: the probability of overlapping timestamps with previous events
    :param p_right: the probability of overlapping timestamps with successive events
    :param max_overlap_left: the maximum number of events that a timestamp can overlap
    :param max_overlap_right: the maximum number of events that a timestamp can overlap
    :param timestamp_key: the xes key for the timestamp
    :param u_timestamp_min_key: the xes key for the minimum value of an uncertain timestamp
    :param u_timestamp_max_key: the xes key for the maximum value of an uncertain timestamp
    :return:
    """

    if p_left > 0.0 or p_right > 0.0:
        for trace in log:
            add_uncertain_timestamp_to_trace_montecarlo(trace, p_left, p_right, max_overlap_left, max_overlap_right, timestamp_key, u_timestamp_min_key, u_timestamp_max_key)


def add_uncertain_timestamp_to_trace_montecarlo(trace, p_left, p_right, max_overlap_left=0, max_overlap_right=0, timestamp_key=DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=DEFAULT_U_TIMESTAMP_MAX_KEY):
    """
    Adds possible activity labels to events in a trace with a certain probability, up to a maximum.

    :param trace: the trace
    :param p_left: the probability of overlapping timestamps with previous events
    :param p_right: the probability of overlapping timestamps with successive events
    :param max_overlap_left: the maximum number of events that a timestamp can overlap
    :param max_overlap_right: the maximum number of events that a timestamp can overlap
    :param timestamp_key: the xes key for the timestamp
    :param u_timestamp_min_key: the xes key for the minimum value of an uncertain timestamp
    :param u_timestamp_max_key: the xes key for the maximum value of an uncertain timestamp
    :return:
    """

    if p_left > 0.0 or p_right > 0.0:
        for i in range(len(trace)):
            steps_left = 0
            steps_right = 0
            while random() < p_left and steps_left < min(max_overlap_left, i):
                steps_left += 1
            while random() < p_right and steps_right < min(max_overlap_right, len(trace) - i - 1):
                steps_right += 1
            # (Partially) supports events that already have uncertainty on timestamps
            # This might cause problems on events for which 'u_timestamp_min_key' <= 'timestamp_key' <= 'u_timestamp_max_key'
            if u_timestamp_max_key in trace[i - steps_left]:
                if u_timestamp_min_key in trace[i]:
                    trace[i][u_timestamp_min_key] = copy(min(trace[i][u_timestamp_min_key], trace[i - steps_left][u_timestamp_max_key]))
                else:
                    trace[i][u_timestamp_min_key] = copy(min(trace[i][timestamp_key], trace[i - steps_left][u_timestamp_max_key]))
            else:
                if u_timestamp_min_key in trace[i]:
                    trace[i][u_timestamp_min_key] = copy(min(trace[i][u_timestamp_min_key], trace[i - steps_left][timestamp_key]))
                else:
                    trace[i][u_timestamp_min_key] = copy(min(trace[i][timestamp_key], trace[i - steps_left][timestamp_key]))
            if u_timestamp_min_key in trace[i + steps_right]:
                if u_timestamp_max_key in trace[i]:
                    trace[i][u_timestamp_max_key] = copy(max(trace[i][u_timestamp_max_key], trace[i + steps_right][u_timestamp_min_key]))
                else:
                    trace[i][u_timestamp_max_key] = copy(max(trace[i][timestamp_key], trace[i + steps_right][u_timestamp_min_key]))
            else:
                if u_timestamp_max_key in trace[i]:
                    trace[i][u_timestamp_max_key] = copy(max(trace[i][u_timestamp_max_key], trace[i + steps_right][timestamp_key]))
                else:
                    trace[i][u_timestamp_max_key] = copy(max(trace[i][timestamp_key], trace[i + steps_right][timestamp_key]))



from random import random, sample

from proved.xes_keys import DEFAULT_U_MISSING_KEY


def add_indeterminate_events_to_log(p, log=None, log_map=None, u_missing_key=DEFAULT_U_MISSING_KEY):
    """
    Turns events in an trace into indeterminate events with a certain probability.

    :param trace: the trace
    :param p: the probability of indeterminate events
    :param u_missing_key: the xes key for indeterminate events
    :return:
    """

    if p > 0.0:

        if log_map is None:
            if log is None:
                raise ValueError('Parameters log and log_map cannot both be None.')
            else:
                log_map = {}
                i = 0
                for trace in log:
                    for j in range(len(trace)):
                        log_map[i] = (trace, j)
                        i += 1

        to_add = max(0, round(len(log_map) * p))
        indices_to_add = sample(frozenset(log_map), to_add)
        for i in indices_to_add:
            trace, j = log_map[i]
            trace[j][u_missing_key] = 1


def add_indeterminate_events_to_log_montecarlo(log, p, u_missing_key=DEFAULT_U_MISSING_KEY):
    """
    Turns events in an event log into indeterminate events with a certain probability.

    :param log: the event log
    :param p: the probability of indeterminate events
    :param u_missing_key: the xes key for indeterminate events
    :return:
    """

    if p > 0.0:
        for trace in log:
            add_indeterminate_events_to_trace_montecarlo(trace, p, u_missing_key)


def add_indeterminate_events_to_trace_montecarlo(trace, p, u_missing_key=DEFAULT_U_MISSING_KEY):
    """
    Turns events in an trace into indeterminate events with a certain probability.

    :param trace: the trace
    :param p: the probability of indeterminate events
    :param u_missing_key: the xes key for indeterminate events
    :return:
    """

    if p > 0.0:
        for event in trace:
            if random() < p:
                event[u_missing_key] = 1



# UNCERTAIN METADATA KEYS

DEFAULT_U_NAME_KEY = 'u:concept:name'
DEFAULT_U_TIMESTAMP_MIN_KEY = 'u:time:timestamp_min'
DEFAULT_U_TIMESTAMP_MAX_KEY = 'u:time:timestamp_max'
DEFAULT_U_MISSING_KEY = 'u:missing'



from random import random, sample

from pm4py.objects.log.util.xes import DEFAULT_NAME_KEY

from proved.xes_keys import DEFAULT_U_NAME_KEY


def add_uncertain_activities_to_log(p, log=None, log_map=None, max_labels_to_add=1, label_set=None, activity_key=DEFAULT_NAME_KEY, u_activity_key=DEFAULT_U_NAME_KEY):
    """
    Adds possible activity labels to events in a trace with a certain probability, up to a maximum.

    :param trace: the trace
    :param p: the probability of indeterminate events
    :param label_set: the list of new possible activity labels
    :param max_labels: the maximum number of uncertain activity labels (unbounded if 0)
    :param activity_key: the xes key for the activity labels
    :param u_activity_key: the xes key for uncertain activity labels
    :return:
    """

    if p > 0.0:

        build_label_set = False
        if label_set is None:
            build_label_set = True

        if log_map is None:
            if log is None:
                raise ValueError('Parameters log and log_map cannot both be None.')
            else:
                log_map = {}
                if build_label_set:
                    label_set = set()
                i = 0
                for trace in log:
                    for j in range(len(trace)):
                        log_map[i] = (trace, j)
                        if build_label_set:
                            label_set.add(trace[j][activity_key])
                        i += 1

        to_alter = max(0, round(len(log_map) * p))
        indices_to_alter = sample(frozenset(log_map), to_alter)
        labels_to_add = min(max_labels_to_add, len(label_set) - 1)
        for i in indices_to_alter:
            trace, j = log_map[i]
            trace[j][u_activity_key] = dict()
            trace[j][u_activity_key]['children'] = {activity_label: 0 for activity_label in [trace[j][activity_key]] + sample(list(label_set - {trace[j][activity_key]}), labels_to_add)}


def add_uncertain_activities_to_log_montecarlo(log, p, max_labels=0, label_set=None, activity_key=DEFAULT_NAME_KEY, u_activity_key=DEFAULT_U_NAME_KEY):
    """
    Adds possible activity labels to events in an event log with a certain probability, up to a maximum.

    :param log: the event log
    :param p: the probability of indeterminate events
    :param label_set: the list of new possible activity labels
    :param max_labels: the maximum number of uncertain activity labels (unbounded if 0)
    :param activity_key: the xes key for the activity labels
    :param u_activity_key: the xes key for uncertain activity labels
    :return:
    """

    if p > 0.0:
        if label_set is None:
            label_set = set()
            for trace in log:
                for event in trace:
                    label_set.add(event[activity_key])
        for trace in log:
            add_uncertain_activities_to_trace_montecarlo(trace, p, label_set, max_labels, activity_key, u_activity_key)


def add_uncertain_activities_to_trace_montecarlo(trace, p, max_labels=0, label_set=None, activity_key=DEFAULT_NAME_KEY, u_activity_key=DEFAULT_U_NAME_KEY):
    """
    Adds possible activity labels to events in a trace with a certain probability, up to a maximum.

    :param trace: the trace
    :param p: the probability of indeterminate events
    :param label_set: the list of new possible activity labels
    :param max_labels: the maximum number of uncertain activity labels (unbounded if 0)
    :param activity_key: the xes key for the activity labels
    :param u_activity_key: the xes key for uncertain activity labels
    :return:
    """

    if p > 0.0:
        if label_set is None:
            label_set = set()
            for event in trace:
                label_set.add(event[activity_key])
        for event in trace:
            to_add = 0
            if max_labels == 0 or max_labels > len(label_set):
                if u_activity_key not in event:
                    max_labels = len(label_set) - 1
                else:
                    max_labels = len(label_set) - len(event[u_activity_key]['children'])
            while random() < p and to_add < max_labels:
                to_add += 1
            if to_add > 0:
                if u_activity_key not in event:
                    event[u_activity_key] = dict()
                    event[u_activity_key]['children'] = {activity_label: 0 for activity_label in [event[activity_key]] + sample(label_set - {event[activity_key]}, to_add)}
                else:
                    event[u_activity_key]['children'].update({activity_label: 0 for activity_label in [event[activity_key]] + sample(label_set - {event[activity_key]}, to_add)})



import operator

from networkx import DiGraph
from pm4py.objects.log.util import xes

import proved.xes_keys as xes_keys


def create_nodes_tuples(trace, activity_key=xes.DEFAULT_NAME_KEY, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=xes_keys.DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=xes_keys.DEFAULT_U_TIMESTAMP_MAX_KEY, u_missing_key=xes_keys.DEFAULT_U_MISSING_KEY, u_activity_key=xes_keys.DEFAULT_U_NAME_KEY):
    # Timestamp type: False is 'minimum', True is 'maximum'
    nodes_tuples = []
    for i, event in enumerate(trace):
        if u_activity_key not in event:
            if u_missing_key not in event:
                new_node = (i, frozenset((event[activity_key],)))
            else:
                new_node = (i, frozenset((event[activity_key], None)))
        else:
            if u_missing_key not in event:
                new_node = (i, frozenset(event[u_activity_key]['children']))
            else:
                new_node = (i, frozenset(tuple(event[u_activity_key]['children']) + (None,)))

        # Fill in the timestamps list
        if u_timestamp_min_key not in event:
            nodes_tuples.append((event[timestamp_key], new_node, False))
            nodes_tuples.append((event[timestamp_key], new_node, True))
        else:
            nodes_tuples.append((event[u_timestamp_min_key], new_node, False))
            nodes_tuples.append((event[u_timestamp_max_key], new_node, True))

    # Sort nodes_tuples by first term of its elements and by type of timestamp
    nodes_tuples.sort(key=operator.itemgetter(0, 2))

    # Returns a tuple of nodes and timestamp types sorted by timestamp
    return tuple((node, timestamp_type) for _, node, timestamp_type in nodes_tuples)


class BehaviorGraph(DiGraph):
    """
    Class representing a behavior graph, a directed acyclic graph showing the precedence relationship between uncertain events.
    For more information refer to:
        Pegoraro, Marco, and Wil MP van der Aalst. "Mining uncertain event data in process mining." 2019 International Conference on Process Mining (ICPM). IEEE, 2019.
    """

    def __init__(self, trace, activity_key=xes.DEFAULT_NAME_KEY, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=xes_keys.DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=xes_keys.DEFAULT_U_TIMESTAMP_MAX_KEY, u_missing_key=xes_keys.DEFAULT_U_MISSING_KEY, u_activity_key=xes_keys.DEFAULT_U_NAME_KEY):
        super().__init__(self)

        node_tuples = create_nodes_tuples(trace, activity_key, timestamp_key, u_timestamp_min_key, u_timestamp_max_key, u_missing_key, u_activity_key)
        edges_list = []

        # Adding the nodes to the graph object
        self.add_nodes_from([node for node, _ in node_tuples])

        # Applies the sweeping algorithm to the sorted list
        for i, node_tuple1 in enumerate(node_tuples):
            node1, type1 = node_tuple1
            if type1 is True:
                for node_tuple2 in node_tuples[i + 1:]:
                    node2, type2 = node_tuple2
                    if type2 is False:
                        edges_list.append((node1, node2))
                    elif (node1, node2) in edges_list:
                        break

        # Adding the edges to the graph object
        self.add_edges_from(edges_list)



from pm4py.objects.log.util.xes import DEFAULT_NAME_KEY, DEFAULT_TIMESTAMP_KEY

from proved.xes_keys import DEFAULT_U_NAME_KEY, DEFAULT_U_TIMESTAMP_MIN_KEY, DEFAULT_U_TIMESTAMP_MAX_KEY, DEFAULT_U_MISSING_KEY
from proved.simulation.bewilderer.add_activities import add_uncertain_activities_to_log
from proved.simulation.bewilderer.add_timestamps import add_uncertain_timestamp_to_log
from proved.simulation.bewilderer.add_indeterminate_events import add_indeterminate_events_to_log


def add_uncertainty(p_a=0.0, p_t=0.0, p_i=0.0, log=None, log_map=None, max_labels_to_add=1, label_set=None, activity_key=DEFAULT_NAME_KEY, u_activity_key=DEFAULT_U_NAME_KEY, timestamp_key=DEFAULT_TIMESTAMP_KEY, u_timestamp_min_key=DEFAULT_U_TIMESTAMP_MIN_KEY, u_timestamp_max_key=DEFAULT_U_TIMESTAMP_MAX_KEY, u_missing_key=DEFAULT_U_MISSING_KEY):
    add_uncertain_activities_to_log(p_a, log, log_map=log_map, max_labels_to_add=max_labels_to_add, label_set=label_set, activity_key=activity_key, u_activity_key=u_activity_key)
    add_uncertain_timestamp_to_log(p_t, log, log_map=log_map, timestamp_key=timestamp_key, u_timestamp_min_key=u_timestamp_min_key, u_timestamp_max_key=u_timestamp_max_key)
    add_indeterminate_events_to_log(p_i, log, log_map=log_map, u_missing_key=u_missing_key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

setup_requirements = []

test_requirements = []

setup(
    author="Marco Pegoraro",
    author_email='pegoraro@pads.rwth-aachen.de',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="PROVED (PRocess mining OVer uncErtain Data) is a collection of tools to perform process mining on uncertain event data.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=['proved', 'process mining', 'data science', 'process discovery', 'conformance checking'],
    name='proved',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/marcopegoraro/proved',
    version='0.1.1',
    zip_safe=False,
    python_requires='>=3.6',
)
